/*
 * Copyright (C) 2022 NHR@FAU, University Erlangen-Nuremberg.
 * All rights reserved.
 * Use of this source code is governed by a MIT-style
 * license that can be found in the LICENSE file.
 */
#include <float.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "parameter.h"
#include "particletracing.h"
#include "progress.h"
#include "timing.h"
#include <mpi.h>

enum VARIANT { SOR = 1, RB, RBA };

int main(int argc, char** argv)
{
    int rank;
    int variant = RB;

    double S, E;
    Parameter params;
    Solver solver;
    ParticleTracer particletracer;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    initParameter(&params);

    if (argc < 2) {
        printf("Usage: %s <configFile>\n", argv[0]);
        exit(EXIT_SUCCESS);
    }

    readParameter(&params, argv[1]);
    if (argc == 3) {
        variant = atoi(argv[2]);
    }

    initSolver(&solver, &params);
    initParticleTracer(&particletracer, &params, &solver);
    if (rank == 0) {
        printParameter(&params);
    }

    printParticleTracerParameters(&particletracer);
    initProgress(solver.te);

    double tau = solver.tau;
    double te  = solver.te;
    double t   = 0.0;

    int (*solver_generic[])() = { solve, solveRB, solveRBA };

    S = getTimeStamp();

    while (t <= te) {
        if (tau > 0.0) {
            computeTimestep(&solver);
        }

        setBoundaryConditions(&solver);
        setSpecialBoundaryCondition(&solver);
        setObjectBoundaryCondition(&solver);
        computeFG(&solver);
        computeRHS(&solver);
        solver_generic[variant - 1](&solver);
        adaptUV(&solver);

        trace(&particletracer, solver.u, solver.v, solver.s, t);

        t += solver.dt;

#ifdef VERBOSE
        if (rank == 0) {
            printf("TIME %f , TIMESTEP %f\n", t, solver.dt);
        }
#else
        printProgress(t);
#endif
    }

    E = getTimeStamp();
    stopProgress();
    if (rank == 0) {
        printf("Solution took %.2fs\n", E - S);
    }
    collectResult(&solver);
    freeParticles(&particletracer);

    MPI_Finalize();
    return EXIT_SUCCESS;
}
